@@ -32,7 +32,7 @@ class CloudPickleFileTests(unittest.TestCase):
         # Empty file
         open(self.tmpfilepath, 'w').close()
         with open(self.tmpfilepath, 'r') as f:
-            self.assertEquals('', pickle.loads(cloudpickle.dumps(f)).read())
+            self.assertEqual('', pickle.loads(cloudpickle.dumps(f)).read())
         os.remove(self.tmpfilepath)
 
     def test_closed_file(self):
@@ -51,7 +51,7 @@ class CloudPickleFileTests(unittest.TestCase):
         # Open for reading
         with open(self.tmpfilepath, 'r') as f:
             new_f = pickle.loads(cloudpickle.dumps(f))
-            self.assertEquals(self.teststring, new_f.read())
+            self.assertEqual(self.teststring, new_f.read())
         os.remove(self.tmpfilepath)
 
     def test_w_mode(self):
@@ -68,7 +68,7 @@ class CloudPickleFileTests(unittest.TestCase):
             f.write(self.teststring)
             f.seek(0)
             new_f = pickle.loads(cloudpickle.dumps(f))
-            self.assertEquals(self.teststring, new_f.read())
+            self.assertEqual(self.teststring, new_f.read())
         os.remove(self.tmpfilepath)
 
     def test_seek(self):
@@ -78,11 +78,11 @@ class CloudPickleFileTests(unittest.TestCase):
             f.seek(4)
             unpickled = pickle.loads(cloudpickle.dumps(f))
             # unpickled StringIO is at position 4
-            self.assertEquals(4, unpickled.tell())
-            self.assertEquals(self.teststring[4:], unpickled.read())
+            self.assertEqual(4, unpickled.tell())
+            self.assertEqual(self.teststring[4:], unpickled.read())
             # but unpickled StringIO also contained the start
             unpickled.seek(0)
-            self.assertEquals(self.teststring, unpickled.read())
+            self.assertEqual(self.teststring, unpickled.read())
         os.remove(self.tmpfilepath)
 
     @pytest.mark.skipif(sys.version_info >= (3,),
@@ -94,12 +94,12 @@ class CloudPickleFileTests(unittest.TestCase):
             f = fp.file
             # FIXME this doesn't work yet: cloudpickle.dumps(fp)
             newfile = pickle.loads(cloudpickle.dumps(f))
-            self.assertEquals(self.teststring, newfile.read())
+            self.assertEqual(self.teststring, newfile.read())
 
     def test_pickling_special_file_handles(self):
         # Warning: if you want to run your tests with nose, add -s option
         for out in sys.stdout, sys.stderr:  # Regression test for SPARK-3415
-            self.assertEquals(out, pickle.loads(cloudpickle.dumps(out)))
+            self.assertEqual(out, pickle.loads(cloudpickle.dumps(out)))
         self.assertRaises(pickle.PicklingError,
                           lambda: cloudpickle.dumps(sys.stdin))
 
